/**********************************************************************
 *  cgi.c -- libcgi
 *
 * Copyright 1994 by the Massachusetts Institute of Technology
 * For copying and distribution information, please see the file
 * <mit-copyright.h>.
 **********************************************************************/
#include "mit-copyright.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include "cgi.h"

char *cgi_query=(char*)0;

int cgi_alloc_form(formp)
     cgi_form *formp;
{
  char *env_var;

  /* zero out everything */
  memset(formp, 0, sizeof(cgi_form));

  /* get query string from environment */
  if (!cgi_query)
    {
      env_var=getenv("REQUEST_METHOD");
      if (!env_var)
	{
	  formp->errcond=1;
	  strcpy(formp->errmsg, "400 REQUEST_METHOD not set.");
	  return(1);
	}
      if (!strcmp(env_var, "GET"))
	{
	  cgi_query= getenv("QUERY_STRING");
	  if (!cgi_query)
	    {
	      formp->errcond=1;
	      strcpy(formp->errmsg,
		     "400 REQUEST_METHOD is GET, but QUERY_STRING is not set.");
	      return(1);
	    }
	}
      else
	{
	  int content_length, total=0;

	  content_length= atoi(getenv("CONTENT_LENGTH"));
	  cgi_query=malloc(content_length+1);
	  if (!cgi_query)
	    {
	      formp->errcond=1;
	      sprintf(formp->errmsg,
		      "503 Couldn't allocate %d bytes of memory.",
		      content_length+1);
	      return(1);
	    }
	  while (total < content_length)
	    total += read(0, cgi_query+total, content_length);
	  cgi_query[content_length]='\0';
	}
    }

  formp->source = cgi_query;
  if (! formp->source)
    {
      formp->errcond=1;
      strcpy(formp->errmsg, "400 Could not read form input.");
      return(1);
    }

  /* decide how much to allocate */
  formp->maxstorage = strlen(formp->source)+1;
  formp->maxfields = formp->maxstorage / 4 + 1;

  /* allocate */
  formp->storage = (char *) malloc(formp->maxstorage);
  if (!formp->storage)
    {
      formp->errcond=1;
      sprintf(formp->errmsg, "503 Couldn't allocate %d bytes of memory.",
	      formp->maxstorage);
      return(1);
    }
  formp->fields = (cgi_field *) malloc(formp->maxfields * sizeof(cgi_field));
  if (!formp->fields)
    {
      free(formp->storage);
      formp->errcond=1;
      sprintf(formp->errmsg, "503 Couldn't allocate %d bytes of memory.",
	      formp->maxfields * sizeof(cgi_field));
      return(1);
    }

  return(0);
}

void
cgi_free_form(formp)
     cgi_form *formp;
{
  if (formp)
    {
      if (formp->storage) free(formp->storage);
      if (formp->fields) free(formp->fields);
    }
  return;
}

/* converts cr/lf pairs to lf, which Unix will handle better */
void
cgi_fix_crlf(string)
     char *string;
{
  char *ptr;

  while ((ptr=strchr(string, '\r')) != NULL)
    {
      if (*(ptr+1) == '\n') strcpy(ptr, ptr+1);
      else *ptr = '\n';
    }
  return;
}

/* returns 1 if string is filled-in filed, 0 if left blank */
int
cgi_nonblank(string)
     char *string;
{
  char *ptr;

  for(ptr=string; *ptr; ptr++)
    if (isgraph((int)(*ptr))) return(1);

  return(0);
}

/*
 * parsing has 4 states:
 * 0: beginning of name
 * 1: after beginning of name
 * 2: beginning of value
 * 3: after beginning of value
 */

int
cgi_parse_form(formp)
     cgi_form *formp;
{
  char *from_ptr, *to_ptr;
  cgi_field *field_ptr;
  char hex_string[3];
  int parse_state=0;
  int hex_int, i;

  /* initialize variables */
  field_ptr=formp->fields;
  formp->nfields=0;
  hex_string[2]='\0';

  /* iterate through all characters in query string */
  for (from_ptr=formp->source, to_ptr=formp->storage;
       *from_ptr != '\0' && to_ptr - formp->storage < formp->maxstorage;
       from_ptr++)
    {
      if (parse_state==0)
	{
	  /* we're at the 'n' in "name=value&name2=value2..." */
	  field_ptr->name = to_ptr;
	  formp->nfields++;
	  parse_state++;
	}
      if (parse_state==2)
	{
	  /* we're at the 'v' in "name=value&name2=value2..." */
	  field_ptr->value = to_ptr;
	  parse_state++;

	  if (from_ptr[0] == '&')
	    {
	      /* we're at the '&' in "name=&..." - value is "" */
	      if (cgi_required(field_ptr->name))
		{
		  formp->errcond=1;
		  sprintf(formp->errmsg,
			  "400 Required field left blank: %-.512s",
			  field_ptr->name);
		  return(1);
		}
	    }
	}
      if (from_ptr[0] == '%' && cgi_ishex((int)from_ptr[1]))
	{
	  /* hex representation like "%2B" in "four=2%2B2" */
	  hex_string[0] = from_ptr[1];
	  hex_string[1] = from_ptr[2];
	  sscanf(hex_string, "%x", &hex_int);
	  *to_ptr++ = (char)hex_int;
	  from_ptr += 1 + cgi_ishex((int)from_ptr[2]);
	  continue;
	}
      if (from_ptr[0] == '+')
	{
	  *to_ptr++ = ' ';
	  continue;
	}
      if (parse_state==1 && from_ptr[0] == '=')
	{
	  /* we're at the '=' in "name=value&name2=value2..." */
	  *to_ptr++ = '\0';
	  parse_state++;
	  continue;
	}	  
      if (from_ptr[0] == '&')
	{
	  *to_ptr++ = '\0';
	  field_ptr++;
	  if (field_ptr > formp->fields + formp->maxfields)
	    {
	      formp->errcond=1;
	      sprintf(formp->errmsg, "400 Exceeded %d fields.",
		      formp->maxfields);
	      return(1);
	    }
	  parse_state=0;
	  continue;
	}
      /* default */
      *to_ptr++ = *from_ptr;
    }

  /* success */
  *to_ptr='\0';
  if (parse_state==2) field_ptr->value = to_ptr;

  /* fix CR/LF; check required fields */
  for (i=0; i<formp->nfields; i++)
    {
      cgi_fix_crlf(formp->fields[i].value);

      if (cgi_required(formp->fields[i].name)
	  && !cgi_nonblank(formp->fields[i].value))
	{
	  formp->errcond=1;
	  sprintf(formp->errmsg,
		  "400 Required field left blank: %-.512s",
		  formp->fields[i].name);
	  return(1);
	}
    }

  return(0);
}

cgi_print_form(formp, outstream)
     cgi_form *formp;
     FILE *outstream;
{
  int i;

  fprintf(outstream,
	  "source=\t%s\nmaxstorage=\t%d\nmaxfields=\t%d\nnfields=\t%d\n",
	  formp->source, formp->maxstorage, formp->maxfields, formp->nfields);
  for (i=0; i<formp->nfields; i++)
    printf("Field %s=\t%s\n", formp->fields[i].name, formp->fields[i].value);
  return;
}

char *
cgi_value(formp, name)
     cgi_form *formp;
     char *name;
{
  int i;

  for (i=0; i<formp->nfields; i++)
    {
      if (!strcmp(formp->fields[i].name, name))
	return(formp->fields[i].value);
    }
  return("");
}

void
cgi_output_value(formp, name, formatstr, outstream)
     cgi_form *formp;
     char *name;
     char *formatstr;
     FILE *outstream;
{
  int i, nfound=0;

  for (i=0; i<formp->nfields; i++)
    {
      if (!strcmp(formp->fields[i].name, name))
	{
	  if (nfound) fputc(' ', outstream);
	  if (formatstr)
	    fprintf(outstream, formatstr, formp->fields[i].value);
	  else
	    fputs(formp->fields[i].value, outstream);
	  nfound++;
	}	  
    }
  return;
}

void
cgi_concat_errno(string)
     char *string;
{
#if HAVE_STRERROR
  strcat(strcat(string, " - "), strerror(errno));
#else
  char errstring[16];

  sprintf(errstring, " (errno: %d)", errno);
  strcat(string, errstring);
#endif
  return;
}

int
cgi_template_fill(formp, templatefile)
     cgi_form *formp;
     char *templatefile;
{
  FILE *tfp;
  char varname[CGI_VARNAME_MAX];
  char formatstr[CGI_VARNAME_MAX];
  int varnamelen=0, inchar, parse_state=0, formatlen=0, substitutions=0;
#if ENABLE_CGIENV
  int cgienv=0;
  char *envval;
#endif /* ENABLE_CGIENV */

  /* open template file */
  tfp = fopen(templatefile, "r");
  if (!tfp)
    {
      formp->errcond=1;
      sprintf(formp->errmsg, "500 Could not open %-.512s", templatefile);
      cgi_concat_errno(formp->errmsg);
      return(1);
    }

  /* open temporary file for filled-in template */
  if (formp->tmpf) rewind(formp->tmpf);
  else formp->tmpf = tmpfile();
  if (!formp->tmpf)
    {
      formp->errcond=1;
      strcpy(formp->errmsg, "503 Could not open temporary file.");
      cgi_concat_errno(formp->errmsg);
      return(1);
    }

  /*
   * parsing states:
   * 0: echo
   * 1: quote
   * 2: variable
   * 3: format
   */
  
  while ((inchar=fgetc(tfp)) != EOF)
    {
      if (parse_state==1)	/* quote */
	{
	  fputc(inchar, formp->tmpf);
	  parse_state=0;
	  continue;
	}
      if (parse_state==0)	/* echo */
	{
	  if (inchar == (int)'\\')
	    {
	      parse_state=1;
	      continue;
	    }
	  if (inchar == (int)'[')
	    {
	      parse_state=2;
	      varnamelen=0;
	      formatlen=0;
	      continue;
	    }
	  fputc(inchar, formp->tmpf);
	  continue;
	}
      if (parse_state==2)	/* variable name */
	{
	  if (inchar == (int)'%')
	    {
	       formatlen=1;
	       strcpy(formatstr, "%");
	       parse_state=3;
	       continue;
	    }

#if ENABLE_CGIENV
	  if (inchar == (int)'$' && varnamelen==0)
	    {
	      cgienv=1;
	      continue;
	    }
#endif /* ENABLE_CGIENV */

	  if (inchar == (int)']')
	    {
	      substitutions++;
	      varname[varnamelen]='\0';
#if ENABLE_CGIENV
	      if (cgienv)
		{
		  cgienv=0;
		  envval=getenv(varname);
		  if (!envval) envval="";
		  if (formatlen > 0)
		    fprintf(formp->tmpf, formatstr, envval);
		  else fputs(envval, formp->tmpf);
		}
	      else
#endif /* ENABLE_CGIENV */
		{
		  if (formatlen > 0)
		    cgi_output_value(formp, varname, formatstr, formp->tmpf);
		  else cgi_output_value(formp, varname, NULL, formp->tmpf);
		}
	      parse_state=0;
	      continue;
	    }
	  varname[varnamelen++]=(char) inchar;
	  if (varnamelen > CGI_VARNAME_MAX)
	    {
	      fclose(tfp);
	      formp->errcond=1;
	      sprintf(formp->errmsg, "403 Variable Name too long (%-.32s...)",
		      varname);
	      return(1);
	    }
	  continue;
	}
      if (parse_state==3)	/* format */
        {
	  if (inchar == (int)',')
	    {
	      formatstr[formatlen]='\0';
	      parse_state=2;
	      continue;
	    }
	  formatstr[formatlen++]=(char) inchar;
	  if (formatlen > CGI_VARNAME_MAX)
	    {
	      fclose(tfp);
	      formp->errcond=1;
	      sprintf(formp->errmsg, "403 Format String too long (%-.32s...)",
		      formatstr);
	      return(1);
	    }
	  continue;
	}
    }
  fclose(tfp);

  /* If there were no variable substitutions in this template,
     someone may be trying to use cgiemail to get at restricted pages */
  if (!substitutions)
    {
      formp->errcond=1;
      sprintf(formp->errmsg, "403 No variable substitutions in \"%-.512s\"",
		      templatefile);
      return(1);
    }

  return 0;
}

int
cgi_mail_template(formp, templatefile)
     cgi_form *formp;
     char *templatefile;
{
  FILE *pipe;
  int retval;
  char buf[BUFSIZ];
  int nbytes;
  char *fromhost, *servername, datebuf[80];
  time_t rightnow;
  struct tm *timeptr;

  /* fill in template */
  retval=cgi_template_fill(formp, templatefile);
  if (retval) return(retval);

  /* open pipe to sendmail */
  if (!(pipe=popen("/usr/lib/sendmail -oi -t", "w")))
    {
      formp->errcond=1;
      strcpy(formp->errmsg, "500 Could not open sendmail pipe.");
      cgi_concat_errno(formp->errmsg);
      return(1);
    }

  /* put "Received: from HOST with HTTP;\n\tdate\n" */
  /* See RFC 822 appendix D. */

  fromhost = getenv("REMOTE_HOST");
  if (!fromhost || !(*fromhost)) fromhost = getenv("REMOTE_ADDR");
  if (fromhost && *fromhost)
    {
      fputs("Received: from ", pipe);
      fputs(fromhost, pipe);
      servername=getenv("SERVER_NAME");
      if (servername && *servername)
	{
	  fputs(" by ", pipe);
	  fputs(servername, pipe);
	}
      fputs(" with HTTP;\n", pipe);

      /* Put date-time (RFC 822 sec. 5, RFC 1123 sec. 5.2.14) */
      /* XXX not compliant if lang is not English or TZ unknown */
      time(&rightnow);
      timeptr=localtime(&rightnow);
      strftime(datebuf, 79, "\t%a, %d %b %Y %H:%M:%S %Z\n", timeptr);
      fputs(datebuf, pipe);
    }

  /* send filled-in template */
  rewind(formp->tmpf);
  do {
    if ((nbytes = fread(buf, sizeof(char), BUFSIZ, formp->tmpf)) > 0)
      fwrite(buf, sizeof(char), nbytes, pipe);
  } while (nbytes == BUFSIZ);

  /* close */
  retval=pclose(pipe);
  if (retval &&
      /* some OS's violate POSIX by returning -1 with ECHILD even
	 though wait() etc. haven't been called.  People who report
	 this spurious error message always say the mail goes through.
	 If mail mysteriously disappears, it's not this program's fault. */
      !(retval==-1 && errno==ECHILD))
    {
      formp->errcond=1;
      sprintf(formp->errmsg, "500 Exit status of sendmail: %d", retval);
      cgi_concat_errno(formp->errmsg);
      return(1);
    }
  return(0);
}

cgi_output_failure(formp, msg)
     cgi_form *formp;
     char *msg;
{
  puts("Content-Type: text/html");
  printf("Status: %s\n\n", formp->errmsg);
  puts("<HEAD><TITLE>Error</TITLE></HEAD>");
  printf("<BODY><H1>Error</H1>%s<P>", msg);
  puts("<SAMP>");
  puts(formp->errmsg);
  puts("</SAMP><P>");
  cgi_output_value(formp, CGI_ADDENDUM, NULL, stdout);
  puts("</BODY>");
  return;
}

void
cgi_ascii2html(instream, outstream)
     FILE *instream, *outstream;
{
  int c;

  while ((c=fgetc(instream)) != EOF)
    switch (c)
      {
      case (int)'&':
	fputs("&amp;", outstream);
	break;
      case (int)'<':
	fputs("&lt;", outstream);
	break;
      case (int)'>':
	fputs("&gt;", outstream);
	break;
      default:
	fputc(c, outstream);
	break;
      }
  return;
}

void
cgi_redirect(url)
     char *url;
{
  printf("Location: %s\n", url);
  puts("Status: 302 CGI redirection\n");
  return;
}

cgi_output_success(formp, msg)
     cgi_form *formp;
     char *msg;
{
  char *success;

  /* maybe use customized success message */
  success=cgi_value(formp, CGI_SUCCESS);
  if (success && *success)
    {
      cgi_redirect(success);
      return;
    }

  /* use generic success message */
  puts("Content-Type: text/html\n");
  puts("<HEAD><TITLE>Success</TITLE></HEAD>");
  printf("<BODY>%s<P><HR>", msg);
  puts("<PRE>");
  rewind(formp->tmpf);
  cgi_ascii2html(formp->tmpf, stdout);
  if (*cgi_value(formp, "debug-source"))
    {
      puts("<HR>");
      puts(formp->source);
      puts("<HR>");
    }
  puts("</PRE><P>");
  cgi_output_value(formp, CGI_ADDENDUM, NULL, stdout);
  puts("</BODY>");

  return;
}

int
cgi_standard_email()
{
  cgi_form form;

  if (cgi_alloc_form(&form) ||
      cgi_parse_form(&form) ||
      cgi_mail_template(&form, getenv("PATH_TRANSLATED")))
    {
      cgi_output_failure(&form, "No email was sent due to an error.");
      cgi_free_form(&form);
      return(1);
    }

  cgi_output_success(&form, "The following email message was sent.");
  cgi_free_form(&form);
  return(0);
}

int
cgi_standard_echo()
{
  cgi_form form;

  if (cgi_alloc_form(&form) ||
      cgi_parse_form(&form) ||
      cgi_template_fill(&form, getenv("PATH_TRANSLATED")))
    {
      cgi_output_failure(&form, "Form was not processed due to an error.");
      cgi_free_form(&form);
      return(1);
    }

  cgi_output_success(&form, "Processed form looks like this:");
  cgi_free_form(&form);
  return(0);
}
